----------------------------------------------------------------
-- KiwiPlates: plates management
----------------------------------------------------------------
local addon = KiwiPlates

----------------------------------------------------------------
-- Caching frequenty used functions
----------------------------------------------------------------
local next = next
local tinsert = tinsert
local tremove = tremove
local tdelete = addon.tdelete

----------------------------------------------------------------
-- Notifications
----------------------------------------------------------------
do
	local Dummy  = function() end
	local Funcs  = {}
	local Notify = setmetatable({}, { __index = function(t,k) t[k] = Dummy; return Dummy end } )
	
	function addon.RegisterNotify(type, func, priority)
		local funcs = Funcs[type]
		if funcs then
			if priority then
				tinsert(funcs, 1, func)
			else
				funcs[#funcs+1] = func
			end	
			Notify[type] = function(...) for i=1,#funcs do (funcs[i])(...) end end 
		else
			Funcs[type] = { func }
			Notify[type] = func
		end
	end

	function addon.UnregisterNotify(type, func)
		local funcs = Funcs[type]
		if funcs then
			tdelete(funcs, func)
			if #funcs==0 then
				Funcs[type]  = nil
				Notify[type] = nil
			elseif #funcs==1 then
				Notify[type] = funcs[1]
			end
		end	
	end

	addon.NotifyF = Funcs	
	addon.Notify  = Notify
end

----------------------------------------------------------------
-- Delayed function execution (on next screen Update)
----------------------------------------------------------------
local runOnce, _runOnce = {}, {}
function addon.RunOnUpdate(func)
	runOnce[#runOnce+1] = func
end

----------------------------------------------------------------
-- Plates managament
----------------------------------------------------------------
local bnot = bit.bnot
local band = bit.band
local bor  = bit.bor
local C_Timer = C_Timer
local UNKNOWNOBJECT = UNKNOWNOBJECT
local Notify      = addon.Notify
local NotifyF     = addon.NotifyF
local RunOnUpdate = addon.RunOnUpdate
local PlateSkinRules = addon.PlateSkinRules
local PlateResetColors = addon.PlateResetColors

local plateCount = 0
local bPlates  = {}
local bvPlates = {}
local kvPlates = {}
local kaPlates = {}
local platesDefaultAlpha = 0.4
local unknownPlates = {}
local Plates = setmetatable({}, { __index = function(t,k) local v={}; t[k]=v; return v end } )
local maxFrameLevel = 0
local moduleBit = 1
local moduleMask = setmetatable( {}, { __index = function(t,k) local m=moduleBit; moduleBit = moduleBit*2; t[k]=m; return m  end } )

-- Plate frame level & alpha management
local function PlateBringToFront(kPlate, bPlate, sender, level)
	local mask = bPlate.layerMask
	if mask==0 or level then
		bPlate.curAlpha = 1
		kPlate:SetAlpha( 1 )
		if level and level>bPlate.curFrameLevel then 
			bPlate.curFrameLevel = level
			kPlate:SetFrameLevel( level ) 
			Notify.PlateFrameLevelUpdate(kPlate)
		end
	end	
	bPlate.layerMask = bor( mask, moduleMask[sender] )			
end

local function PlateSendToBack(kPlate, bPlate, sender, level)
	local mask = bPlate.layerMask
	if mask~=0 then
		mask = band( mask, bnot(moduleMask[sender]) )
		if mask==0 then
			bPlate.curAlpha = platesDefaultAlpha
			kPlate:SetAlpha( platesDefaultAlpha )
			bPlate.curFrameLevel = bPlate.defFrameLevel
			kPlate:SetFrameLevel( bPlate.curFrameLevel )
			Notify.PlateFrameLevelUpdate(kPlate)
		elseif level==bPlate.curFrameLevel then	
			bPlate.curFrameLevel = bPlate.defFrameLevel
			kPlate:SetFrameLevel( bPlate.curFrameLevel )
			Notify.PlateFrameLevelUpdate(kPlate)
		end	
		bPlate.layerMask = mask
	end	
end

local function PlatesSetAlpha(alpha)
	platesDefaultAlpha = alpha or 0.4
	for i=#bvPlates,1,-1 do
		local bPlate = bvPlates[i]
		if bPlate.layerMask==0 then
			bPlate.curAlpha = platesDefaultAlpha
			bPlate.kPlate:SetAlpha( platesDefaultAlpha )
		end
	end
end

--- Plate Factory
local function PlateFactoryGet(skinName)
	local kPlate = tremove(Plates[skinName])
	if not kPlate then
		local skin = addon.skins[skinName] or addon.skins['Basic.TextOnly']
		kPlate = CreateFrame( 'Frame' , nil, WorldFrame)
		kPlate.db = skin
		kPlate.skinName = skinName
		kaPlates[#kaPlates+1] = kPlate
		Notify.PlateCreate(kPlate)
	end
	return kPlate
end

local function PlateFactoryPut(kPlate)
	tinsert( Plates[kPlate.skinName], kPlate )
	PlateResetColors(kPlate)
end

-- Tracking posible unknown entities
local function UpdateUnknownPlates()
	for bPlate in next,unknownPlates do 
		local kPlate = bPlate.kPlate
		if kPlate then
			local name = bPlate.bNameText:GetText()
			if name ~= UNKNOWNOBJECT then
				unknownPlates[bPlate] = nil
				bPlate.vName = name
				Notify.bPlateShow(bPlate)
				if not (PlateSkinRules.names and addon.ReSkinPlate(bPlate)) then
					Notify.PlateShow(kPlate,bPlate)
				end	
			end
		else
			unknownPlates[bPlate] = nil
		end	
	end
	if next(unknownPlates) then
		C_Timer.After(1,UpdateUnknownPlates)
	end
end

local function GetPlateName(bPlate)
	local name = bPlate.bNameText:GetText()
	if name == UNKNOWNOBJECT then 
		local empty = not next(unknownPlates)
		unknownPlates[bPlate] = true
		if empty then C_Timer.After(1,UpdateUnknownPlates) end
	end
	return name
end

-- Blizzard plate was shown
local function ShowPlate(bPlate)
	RunOnUpdate( function()
		if bPlate:IsShown() then
			tinsert( bvPlates, bPlate )		
			bPlate.layerMask = 0			
			bPlate.curFrameLevel = bPlate.defFrameLevel
			bPlate.curAlpha = platesDefaultAlpha
			bPlate.vName = GetPlateName(bPlate)
			Notify.bPlateShow(bPlate)
			local kPlate = PlateFactoryGet( addon.PlateGetSkinName(bPlate) )
			tinsert( kvPlates, kPlate )			
			kPlate.bPlate = bPlate
			bPlate.kPlate = kPlate
			kPlate:SetFrameLevel( bPlate.curFrameLevel )
			kPlate:SetAlpha( bPlate.curAlpha )
			Notify.PlateShow(kPlate,bPlate)						
			bPlate.bEliteIcon:SetAlpha(0)
			bPlate.bLevelText:Hide()
		end	
	end )	
end

-- Blizzard plate was hidden
local function HidePlate(bPlate)
	local kPlate = bPlate.kPlate
	if kPlate then
		kPlate:Hide()
		Notify.bPlateHide( bPlate )
		Notify.PlateHide( kPlate )
		PlateFactoryPut(kPlate)
		kPlate.bPlate = nil		
		bPlate.kPlate = nil
		bPlate.layerMask = 0
		tdelete( bvPlates, bPlate )
		tdelete( kvPlates, kPlate )
		unknownPlates[bPlate] = nil
	end	
end

-- Plates reskin
local function ReSkinPlate(bPlate, force)
	local kPlate   = bPlate.kPlate
	local skinName = addon.PlateGetSkinName(bPlate)
	if skinName ~= kPlate.skinName or force then
		kPlate:Hide()
		Notify.PlateHide( kPlate )		
		PlateFactoryPut(kPlate)
		kPlate.bPlate = nil		
		tdelete( kvPlates, kPlate )
		kPlate = PlateFactoryGet(skinName)
		kPlate.bPlate = bPlate
		bPlate.kPlate = kPlate
		kvPlates[#kvPlates+1] = kPlate
		kPlate:SetFrameLevel( bPlate.curFrameLevel )
		kPlate:SetAlpha( bPlate.curAlpha )
		Notify.PlateShow(kPlate,bPlate) 		
		return true
	end	
end

local function ReSkinPlates(force)
	for i=#bvPlates,1,-1 do 
		ReSkinPlate( bvPlates[i] ,force )
	end
end

-- Refresh statuses
local function UpdateStatuses()
	for i=#bvPlates,1,-1 do 
		Notify.bPlateShow(bvPlates[i])
	end
end

-- New Blizzard plate detected
local function InitPlate(bPlate)
	local nam = bPlate.NameContainer
	local art = bPlate.ArtContainer
	nam.NameText:Hide()	
	art.AbsorbBar.Overlay:SetTexture(nil)
    art.Border:SetTexture(nil)
    art.Highlight:SetTexture(nil)
    art.HighLevelIcon:SetTexture(nil)
    art.CastBarFrameShield:SetTexture(nil)
    art.CastBarBorder:SetTexture(nil)
    art.AggroWarningTexture:SetTexture(nil)
    art.CastBarSpellIcon:SetSize(.01,.01)
    art.CastBarTextBG:SetTexture(nil)
	art.HealthBar:Hide()
	art.CastBar:Hide()
    art.CastBarText:Hide()	
	art.AbsorbBar:SetStatusBarTexture(nil)
    art.CastBar:SetStatusBarTexture(nil)
	art.LevelText:SetWidth( 0.01 )
	art.LevelText:Hide()
	bPlate.bCastBar   = art.CastBar
	bPlate.bNameText  = nam.NameText	
	bPlate.bHealthBar = art.HealthBar
	bPlate.bHighlight = art.Highlight
	bPlate.bBossIcon  = art.HighLevelIcon
	bPlate.bEliteIcon = art.EliteIcon
	bPlate.bLevelText = art.LevelText
	bPlate.bRaidTargetIcon = art.RaidTargetIcon
	bPlate.bAggroWarningTexture = art.AggroWarningTexture
	bPlate.bSpellIcon = art.CastBarSpellIcon
	bPlate.bSpellText = art.CastBarText	
	bPlate.bShield = art.CastBarFrameShield
	bPlate.defFrameLevel = maxFrameLevel
	if maxFrameLevel<126 then maxFrameLevel = maxFrameLevel + 1 end
	bPlates[#bPlates+1] = bPlate
	bPlate:HookScript( 'OnShow', ShowPlate )
	bPlate:HookScript( 'OnHide', HidePlate )
	bPlate.kInit = true
	Notify.bPlateCreate(bPlate)	
    if bPlate:IsShown() then ShowPlate(bPlate) end
end

----------------------------------------------------------------
-- Blizzard Plates tracking and Timer Events
----------------------------------------------------------------
do
	local elapsed = 0
	local wipe = wipe	
	local select = select
	local strfind = strfind
    local WorldFrame = WorldFrame
	local GetChildren = WorldFrame.GetChildren
	local GetNumChildren = WorldFrame.GetNumChildren
	local function Update(count, ...)
		for i = 1, count do
			local bPlate = select(i,...)
			if (not bPlate.kInit) and bPlate.ArtContainer and strfind(bPlate:GetName() or '','^NamePlate%d') then
				InitPlate(bPlate)
			end
		end
	end
	function addon:PlatesRun(frame)
		addon.uiScale = UIParent:GetEffectiveScale()
		frame:HookScript( 'OnUpdate', function(self, e)
			-- Search new nameplates
			local count = GetNumChildren(WorldFrame)
			if count ~= plateCount then
				Update( count, GetChildren(WorldFrame) )
				plateCount = count
			end
			-- RunOnce functions
			if #runOnce>0 then
				wipe(_runOnce)
				runOnce, _runOnce = _runOnce, runOnce
				for i=1,#_runOnce do _runOnce[i]() end
			end
			-- General purpose timer event, every 1/10 seconds
			elapsed = elapsed - e
			if elapsed<=0 then
				elapsed = 0.1
				if NotifyF.Update then Notify.Update() end
			end
		end )
		self.PlatesRun = nil
	end	
end

-------------------------------------------------------
--- Published methods and variables
-------------------------------------------------------
addon.UpdateStatuses    = UpdateStatuses
addon.ReSkinPlate       = ReSkinPlate
addon.ReSkinPlates      = ReSkinPlates
addon.PlateSendToBack   = PlateSendToBack
addon.PlateBringToFront = PlateBringToFront
addon.PlatesSetAlpha    = PlatesSetAlpha

addon.kaPlates = kaPlates -- all kPlates
addon.kvPlates = kvPlates -- visible kPlates
addon.bvPlates = bvPlates -- blizzard visible plates
addon.bPlates  = bPlates  -- all blizzard plates
